import argparse
import math
import random
from pathlib import Path
from typing import Any, Dict, List, Tuple


# Function for generating random centers based on the configuration
def generate_centers(
    number_centers: int, max_value: float, dimensions: int
) -> List[List[float]]:
    centers = []
    for _ in range(number_centers):
        # Create a center with random coordinates in each dimension
        center = [random.uniform(0, max_value) for _ in range(dimensions)]
        centers.append(center)

    return centers


# Function for generating points around a given center based on the configuration
def generate_points_around_center(
    center: List[float],
    min_points_per_cluster: int,
    max_points_per_cluster: int,
    min_cluster_radius: float,
    max_cluster_radius: float,
    max_value: float,
    dimensions: int,
) -> List[List[float]]:
    # Determine the number of points in the cluster and the cluster radius
    number_points = random.randint(min_points_per_cluster, max_points_per_cluster)
    cluster_radius = random.uniform(min_cluster_radius, max_cluster_radius)
    points = []
    count = 0
    while count < number_points:
        # Generate a random radius and random angle for each dimension
        radius = random.uniform(0, cluster_radius)
        angles = [random.uniform(0, 2 * math.pi) for _ in range(dimensions)]
        point = []
        valid_point = True
        for i in range(dimensions):
            # Calculate the coordinate using the radius and angle
            coord = radius * math.cos(angles[i]) + center[i]
            # Check if the coordinate is within the valid limits
            if coord < 0 or coord > max_value:
                valid_point = False
                break
            point.append(coord)
        if valid_point:
            # Add the point to the list if it is valid
            points.append(point)
            count += 1

    return points


# Function for generating clusters based on the configuration
def generate_clusters(
    number_centers: int,
    max_value: float,
    min_points_per_cluster: int,
    max_points_per_cluster: int,
    min_cluster_radius: float,
    max_cluster_radius: float,
    dimensions: int,
) -> Tuple[List[List[float]], List[List[float]]]:
    # Generate the centers
    centers = generate_centers(number_centers, max_value, dimensions)
    points: List[List[float]] = []
    for center in centers:
        # Generate points around each center
        points = points + generate_points_around_center(
            center=center,
            min_points_per_cluster=min_points_per_cluster,
            max_points_per_cluster=max_points_per_cluster,
            min_cluster_radius=min_cluster_radius,
            max_cluster_radius=max_cluster_radius,
            max_value=max_value,
            dimensions=dimensions,
        )

    return centers, points


# Function for writing the generated points to a CSV file
def write_clusters_to_file(output_path: Path, points: List[List[float]]) -> None:
    with output_path.open("r") as file:
        for point in points:
            # Use join to concatenate the coordinates with commas
            line = ",".join(map(str, point))
            file.write(line + "\n")


# Function for handling command line arguments
def handle_arguments() -> Dict[str, Any]:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "-mv",
        "--max-value",
        type=float,
        help="maximum value for any dimension",
        default=10,
    )
    parser.add_argument(
        "-d",
        "--dimensions",
        type=int,
        help="number of dimensions for the points and centers",
        default=2,
    )
    parser.add_argument(
        "-n",
        "--number-centers",
        type=int,
        help="number of centers/clusters that will be generated",
        default=3,
    )
    parser.add_argument(
        "-minp",
        "--min-points-per-cluster",
        type=int,
        help="minimum number of points per cluster (center excluded)",
        default=25,
    )
    parser.add_argument(
        "-maxp",
        "--max-points-per-cluster",
        type=int,
        help="maximum number of points per cluster (center excluded)",
        default=25,
    )
    parser.add_argument(
        "-minr",
        "--min-cluster-radius",
        type=float,
        help="minimum radius of generated clusters",
        default=0.5,
    )
    parser.add_argument(
        "-maxr",
        "--max-cluster-radius",
        type=float,
        help="minimum radius of generated clusters",
        default=2,
    )
    parser.add_argument("-k", "--k", type=int, help="number of clusters", default=3)
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        help="output file for the generated data",
    )
    args = parser.parse_args()

    # Validations for the arguments
    assert args.dimensions > 0
    assert args.max_value > 0
    assert args.number_centers > 0
    assert args.min_points_per_cluster > 0
    assert args.min_points_per_cluster <= args.max_points_per_cluster
    assert args.min_cluster_radius > 0
    assert args.min_cluster_radius <= args.max_cluster_radius

    return vars(args)


# Main function for generating the data
def generate_data(config: Dict[str, Any]) -> None:
    # Generate the clusters
    centers, points = generate_clusters(
        number_centers=config["number_centers"],
        max_value=config["max_value"],
        min_points_per_cluster=config["min_points_per_cluster"],
        max_points_per_cluster=config["max_points_per_cluster"],
        min_cluster_radius=config["min_cluster_radius"],
        max_cluster_radius=config["max_cluster_radius"],
        dimensions=config["dimensions"],
    )
    # Write the generated data to a file
    write_clusters_to_file(
        config["output_path"].parent / (config["output_path"].name + ".data"), points
    )
    write_clusters_to_file(
        config["output_path"].parent / (config["output_path"].name + ".centers"),
        centers,
    )


def main() -> None:
    random.seed(1234)
    config = handle_arguments()
    generate_data(config)


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import Any, List, Optional, Sequence, Tuple

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.patches import Circle


def plot_2d_ax(
    points: np.ndarray,
    ax: plt.Axes,
    clusters: Optional[Sequence[int]] = None,
    centers: Optional[Sequence[Sequence[float]]] = None,
    radii: Optional[Sequence[float]] = None,
    title: Optional[str] = None,
    edge_color: str = "black",
    cmap: str = "plasma",
) -> None:
    ax.scatter(points[:, 0], points[:, 1], c=clusters, s=20, cmap=cmap)
    _centers = np.array(centers)
    ax.scatter(_centers[:, 0], _centers[:, 1], c="red", s=50, alpha=0.5, marker="X")

    if centers is not None and radii is not None:
        for i, ((x, y), radius) in enumerate(zip(centers, radii)):
            circle = Circle(
                (x, y),
                radius,
                fill=False,
                edgecolor=edge_color,
                linestyle="--",
                alpha=0.75,
            )
            ax.add_patch(circle)

            if clusters is not None:
                _clusters = np.array(clusters)
                points_in_cluster = points[_clusters == i]
                furthest_point = points_in_cluster[
                    np.argmax(np.linalg.norm(points_in_cluster - centers[i], axis=1))
                ]
                ax.add_line(
                    plt.Line2D(
                        [centers[i][0], furthest_point[0]],
                        [centers[i][1], furthest_point[1]],
                        color=edge_color,
                        linestyle="--",
                    )
                )
    if title is not None:
        ax.set_title(title)
    ax.set_xlabel("Feature 1")
    ax.set_ylabel("Feature 2")

    ax.set_aspect("equal")


def plot_3d_ax(
    points: np.ndarray,
    ax: plt.Axes,
    clusters: Optional[Sequence[int]] = None,
    centers: Optional[Sequence[Sequence[float]]] = None,
    radii: Optional[Sequence[float]] = None,
    title: Optional[str] = None,
    edge_color: str = "black",
    cmap: str = "plasma",
) -> None:
    ax.scatter(  # type: ignore
        points[:, 0], points[:, 1], points[:, 2], c=clusters, s=20, cmap=cmap
    )
    _centers = np.array(centers)
    ax.scatter(  # type: ignore
        _centers[:, 0],
        _centers[:, 1],
        _centers[:, 2],
        c="red",
        s=50,
        alpha=0.5,
        marker="X",
    )

    if centers is not None and radii is not None:
        for i, ((c_x, c_y, c_z), radius) in enumerate(zip(centers, radii)):
            u, v = np.mgrid[0 : 2 * np.pi : 20j, 0 : np.pi : 10j]  # type: ignore
            x = c_x + radius * np.cos(u) * np.sin(v)
            y = c_y + radius * np.sin(u) * np.sin(v)
            z = c_z + radius * np.cos(v)
            ax.plot_wireframe(  # type: ignore
                x, y, z, linestyle="--", color=edge_color, alpha=0.3
            )

            if clusters is not None:
                _clusters = np.array(clusters)
                points_in_cluster = points[_clusters == i]
                furthest_point = points_in_cluster[
                    np.argmax(np.linalg.norm(points_in_cluster - centers[i], axis=1))
                ]

                ax.plot(
                    [_centers[i, 0], furthest_point[0]],
                    [_centers[i, 1], furthest_point[1]],
                    [_centers[i, 2], furthest_point[2]],
                    color=edge_color,
                    linestyle="--",
                )
    if title is not None:
        ax.set_title(title)

    ax.set_xlabel("Feature 1")
    ax.set_ylabel("Feature 2")
    ax.set_zlabel("Feature 3")  # type: ignore

    ax.set_aspect("equal")


def plot_ax(points: np.ndarray, *args: Any, **kwargs: Any) -> None:
    if len(points[0]) == 2:
        plot_2d_ax(points, *args, **kwargs)
    elif len(points[0]) == 3:
        plot_3d_ax(points, *args, **kwargs)
    else:
        raise ValueError("Only 2D and 3D data is supported for plotting.")


def get_projection(dim: int) -> Optional[str]:
    if dim == 2:
        return None
    elif dim == 3:
        return "3d"
    else:
        raise ValueError("Only 2D and 3D data is supported for plotting.")


def plot_result(
    points: Sequence[Sequence[float]],
    clusters: Optional[Sequence[int]] = None,
    centers: Optional[Sequence[Sequence[float]]] = None,
    radii: Optional[Sequence[float]] = None,
    title: Optional[str] = None,
    *,
    cmap: str = "plasma",
    output_path: Optional[Path] = None,
    transparent: bool = True,
    show: bool = True,
    close: bool = True,
) -> Tuple[plt.Figure, plt.Axes]:
    fig, ax = plt.subplots(
        figsize=(10, 10), subplot_kw=dict(projection=get_projection(len(points[0])))
    )

    plot_ax(np.array(points), ax, clusters, centers, radii, title, cmap=cmap)

    if output_path is not None:
        Path(output_path).parent.mkdir(parents=True, exist_ok=True)
        plt.savefig(
            output_path,
            dpi=300,
            bbox_inches="tight",
            transparent=transparent,
        )
    if show:
        plt.show()

    if close:
        plt.close()

    return fig, ax


def plot_multiple_results(
    points: Sequence[Sequence[float]],
    clusters: Optional[Sequence[Optional[Sequence[int]]]] = None,
    centers: Optional[Sequence[Optional[Sequence[Sequence[float]]]]] = None,
    radii: Optional[Sequence[Optional[Sequence[float]]]] = None,
    title: Optional[Optional[Sequence[str]]] = None,
    *,
    cmap: str = "plasma",
    output_path: Optional[Path] = None,
    transparent: bool = True,
    show: bool = True,
    close: bool = True,
) -> Tuple[plt.Figure, List[plt.Axes]]:
    proj = get_projection(len(points[0]))

    if clusters is not None:
        ll = len(clusters)
    elif centers is not None:
        ll = len(centers)
    elif radii is not None:
        ll = len(radii)
    else:
        ll = 1

    if ll == 1:
        fig, ax = plt.subplots(1, 1, figsize=(10, 10), subplot_kw=dict(projection=proj))
        axs: List[Axes] = [ax]
    else:
        fig, axs = plt.subplots(  # type: ignore
            1,
            ll,
            figsize=(10 * ll, 10),
            subplot_kw=dict(projection=proj),
        )

    if clusters is None:
        clusters = [None] * ll

    if centers is None:
        centers = [None] * ll

    if radii is None:
        radii = [None] * ll

    if title is None:
        title = [f"Plot {i}" for i in range(ll)]

    for params in zip(axs, clusters, centers, radii, title):
        plot_ax(np.array(points), *params, cmap=cmap)

    if output_path is not None:
        Path(output_path).parent.mkdir(parents=True, exist_ok=True)
        plt.savefig(
            output_path,
            dpi=300,
            bbox_inches="tight",
            transparent=transparent,
        )

    if show:
        plt.show()

    if close:
        plt.close()

    return fig, axs



import os
import platform
import subprocess
import warnings
from typing import Any, Dict, List

from setuptools import Extension
from setuptools.command.build_ext import build_ext
from setuptools.errors import CompileError

extension = Extension(
    name="kmsr._core",
    sources=[
        "kmsr/_core.cpp",
        "kmsr/cpp/ball.cpp",
        "kmsr/cpp/cluster.cpp",
        "kmsr/cpp/gonzalez.cpp",
        "kmsr/cpp/heuristic.cpp",
        "kmsr/cpp/util.cpp",
        "kmsr/cpp/k_MSR.cpp",
        "kmsr/cpp/point.cpp",
        "kmsr/cpp/welzl.cpp",
        "kmsr/cpp/yildirim.cpp",
    ],
    include_dirs=["kmsr"],
)


def check_openmp_support() -> List[str]:
    if platform.system() == "Windows":
        return []

    openmp_test_code = """
    #include <omp.h>
    #include <stdio.h>
    int main() {
        int nthreads;
        #pragma omp parallel
        {
            nthreads = omp_get_num_threads();
        }
        printf("Number of threads = %d\\n", nthreads);
        return 0;
    }
    """

    with open("test_openmp.c", "w") as f:
        f.write(openmp_test_code)

    try:
        if platform.system() == "Darwin":
            cmd_params = [
                "-Xclang",
                "-fopenmp",
                "-I/usr/local/opt/libomp/include",
                "-L/usr/local/opt/libomp/lib",
                "-lomp",
            ]
        else:
            cmd_params = ["-fopenmp"]
        # Try to compile the code with OpenMP support
        result = subprocess.run(
            ["gcc"] + cmd_params + ["test_openmp.c", "-o", "test_openmp"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        if result.returncode != 0:
            return []

        # Run the compiled program
        result = subprocess.run(
            ["./test_openmp"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        if result.returncode == 0:
            return cmd_params
        else:
            return []
    finally:
        os.remove("test_openmp.c")
        if os.path.exists("test_openmp"):
            os.remove("test_openmp")


class BuildExt(build_ext):
    """A custom build extension for adding -stdlib arguments for clang++."""

    def build_extensions(self) -> None:
        openmp_params = check_openmp_support()

        # '-std=c++11' is added to `extra_compile_args` so the code can compile
        # with clang++. This works across compilers (ignored by MSVC).
        for extension in self.extensions:
            extension.extra_compile_args.append("-std=c++11")
            if len(openmp_params) == 0:
                warnings.warn(
                    "\x1b[31;20m OpenMP is not installed on this system. "
                    "Please install it to have all the benefits from the program.\x1b[0m"
                )
            else:
                for param in openmp_params:
                    extension.extra_compile_args.append(param)

                if platform.system() == "Linux":
                    extension.extra_link_args.append("-fopenmp")
                    extension.extra_link_args.append("-lgomp")
        try:
            build_ext.build_extensions(self)
        except CompileError:
            # Workaround Issue #2.
            # '-stdlib=libc++' is added to `extra_compile_args` and `extra_link_args`
            # so the code can compile on macOS with Anaconda.
            for extension in self.extensions:
                extension.extra_compile_args.append("-stdlib=libc++")
                extension.extra_link_args.append("-stdlib=libc++")
            build_ext.build_extensions(self)


def build(setup_kwargs: Dict[str, Any]) -> None:
    setup_kwargs.update(
        {"ext_modules": [extension], "cmdclass": {"build_ext": BuildExt}}
    )



import ctypes
from time import time
from typing import Any, Optional, Sequence, Type

import numpy as np
from sklearn.base import BaseEstimator, ClassNamePrefixFeaturesOutMixin, ClusterMixin
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import validate_data

import kmsr._core  # type: ignore

_DLL = ctypes.cdll.LoadLibrary(kmsr._core.__file__)


class KMSR(BaseEstimator, ClusterMixin, ClassNamePrefixFeaturesOutMixin):
    NOT_FITTED_ERROR = (
        "This KMSR instance is not fitted yet. "
        "Call 'fit' with appropriate arguments before using this estimator."
    )

    def __init__(
        self,
        n_clusters: int = 3,
        algorithm: str = "auto",
        epsilon: float = 0.5,
        n_u: int = 1000,
        n_test_radii: int = 10,
        random_state: Optional[int] = None,
    ) -> None:
        self.random_state = random_state
        self.n_clusters = n_clusters
        self.epsilon = epsilon
        self.algorithm = algorithm
        self.n_u = n_u
        self.n_test_radii = n_test_radii

    @property
    def inertia_(self) -> float:
        if not hasattr(self, "_inertia"):
            raise NotFittedError(self.NOT_FITTED_ERROR)
        return self._inertia

    @property
    def labels_(self) -> np.ndarray:
        if not hasattr(self, "_labels"):
            raise NotFittedError(self.NOT_FITTED_ERROR)
        return self._labels

    @property
    def cluster_centers_(self) -> np.ndarray:
        if not hasattr(self, "_cluster_centers"):
            raise NotFittedError(self.NOT_FITTED_ERROR)
        return self._cluster_centers

    @property
    def cluster_radii_(self) -> np.ndarray:
        if not hasattr(self, "_cluster_radii"):
            raise NotFittedError(self.NOT_FITTED_ERROR)
        return self._cluster_radii

    @property
    def real_n_clusters_(self) -> int:
        if not hasattr(self, "_real_n_clusters"):
            raise NotFittedError(self.NOT_FITTED_ERROR)
        return self._real_n_clusters

    def fit(
        self,
        X: Sequence[Sequence[float]],
        y: Any = None,
        sample_weight: Optional[Sequence[float]] = None,
    ) -> "KMSR":
        if sample_weight is not None:
            raise NotImplementedError("sample_weight is not supported")

        return self._fit(X)

    def _fit(
        self,
        X: Sequence[Sequence[float]],
    ) -> "KMSR":
        self._validate_params()

        _algorithm = (
            "fpt-heuristic" if self.algorithm == "auto" else self.algorithm.lower()
        )
        _seed = int(time()) if self.random_state is None else self.random_state
        _X = validate_data(
            self,
            X,
            accept_sparse=False,
            dtype=np.float64,
            order="C",
            accept_large_sparse=False,
            copy=False,
            reset=True,
        )

        n_samples = _X.shape[0]

        _X = np.ascontiguousarray(_X, dtype=np.float64)
        c_array = _X.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

        c_n = ctypes.c_int(n_samples)
        c_features = ctypes.c_int(self.n_features_in_)
        c_clusters = ctypes.c_int(self.n_clusters)
        found_clusters = ctypes.c_int()
        c_seed = ctypes.c_int(_seed)

        labels = np.empty(n_samples, dtype=np.int32, order="C")
        centers = np.empty(
            (self.n_clusters, self.n_features_in_), dtype=np.float64, order="C"
        )
        radii = np.empty(self.n_clusters, dtype=np.float64, order="C")

        c_labels = labels.ctypes.data_as(ctypes.POINTER(ctypes.c_int))
        c_centers = centers.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
        c_radii = radii.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

        # Define common argtypes for the wrappers
        common_argtypes: list[Type[ctypes._CData]] = [
            ctypes.POINTER(ctypes.c_double),  # X
            ctypes.c_int,  # n_samples
            ctypes.c_int,  # n_features
            ctypes.c_int,  # n_clusters
            ctypes.POINTER(ctypes.c_int),  # found_clusters
            ctypes.POINTER(ctypes.c_int),  # labels
            ctypes.POINTER(ctypes.c_double),  # centers
            ctypes.POINTER(ctypes.c_double),  # radii
            ctypes.c_int,  # seed
        ]

        if _algorithm == "fpt-heuristic":
            c_epsilon = ctypes.c_double(self.epsilon)
            c_u = ctypes.c_int(self.n_u)
            c_num_radii = ctypes.c_int(self.n_test_radii)

            schmidt_argtypes = (
                common_argtypes[:4]
                + [ctypes.c_double, ctypes.c_int, ctypes.c_int]
                + common_argtypes[4:]
            )

            _DLL.schmidt_wrapper.argtypes = schmidt_argtypes  # type: ignore[assignment]

            _DLL.schmidt_wrapper.restype = ctypes.c_double

            self._inertia: float = _DLL.schmidt_wrapper(
                c_array,
                c_n,
                c_features,
                c_clusters,
                c_epsilon,
                c_u,
                c_num_radii,
                ctypes.byref(found_clusters),
                c_labels,
                c_centers,
                c_radii,
                c_seed,
            )
        else:
            if _algorithm == "heuristic":
                wrapper_function = _DLL.heuristic_wrapper
            elif _algorithm == "gonzalez":
                wrapper_function = _DLL.gonzalez_wrapper
            elif _algorithm == "kmeans":
                wrapper_function = _DLL.kmeans_wrapper
            else:
                raise ValueError(f"Invalid algorithm: {self.algorithm}")

            wrapper_function.argtypes = common_argtypes  # type: ignore[assignment]
            wrapper_function.restype = ctypes.c_double

            self._inertia = wrapper_function(
                c_array,
                c_n,
                c_features,
                c_clusters,
                ctypes.byref(found_clusters),
                c_labels,
                c_centers,
                c_radii,
                c_seed,
            )

        self._real_n_clusters = found_clusters.value

        self._cluster_centers = centers
        self._cluster_radii = radii

        # Crop the centers and the radii in case the algorithm found less clusters
        self._cluster_centers = self.cluster_centers_[
            : self._real_n_clusters
        ]  # type: ignore[assignment]
        self._cluster_radii = self.cluster_radii_[
            : self._real_n_clusters
        ]  # type: ignore[assignment]

        self._labels = labels

        return self

    def _validate_params(self) -> None:
        if not isinstance(self.n_clusters, (int, np.integer)):
            raise TypeError(
                f"n_clusters must be an integer, got {type(self.n_clusters).__name__}"
            )
        if self.n_clusters < 1:
            raise ValueError(f"n_clusters must be >= 1, got {self.n_clusters}")

        valid_algos = {"auto", "fpt-heuristic", "heuristic", "gonzalez", "kmeans"}
        if self.algorithm.lower() not in valid_algos:
            raise ValueError(
                f"algorithm must be one of {valid_algos}; got {self.algorithm}"
            )

        if not isinstance(self.epsilon, (float, np.floating)):
            raise TypeError(
                f"epsilon must be a float, got {type(self.epsilon).__name__}"
            )
        if self.epsilon < 0:
            raise ValueError(f"epsilon must be >= 0; got {self.epsilon}")

        if not isinstance(self.n_u, (int, np.integer)):
            raise TypeError(f"n_u must be an integer," f"got {type(self.n_u).__name__}")
        if self.n_u < 1:
            raise ValueError(f"n_u must be >= 1, got {self.n_u}")
        if not isinstance(self.n_test_radii, (int, np.integer)):
            raise TypeError(
                f"n_test_radii must be an integer,"
                f"got {type(self.n_test_radii).__name__}"
            )
        if self.n_test_radii < 1:
            raise ValueError(f"n_test_radii must be >= 1, got {self.n_test_radii}")

        if self.random_state is not None:
            if not isinstance(self.random_state, (int, np.integer)):
                raise TypeError(
                    f"random_state must be None or an integer,"
                    f"got {type(self.random_state).__name__}"
                )
            if self.random_state < 0:
                raise ValueError(f"random_state must be >= 0, got {self.random_state}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kmsr']

package_data = \
{'': ['*'], 'kmsr': ['cpp/*', 'header/*']}

install_requires = \
['scikit-learn>=1.6.1,<2.0.0']

setup_kwargs = {
    'name': 'kmsr',
    'version': '0.1.4',
    'description': 'kMSR provides a selection of algorithms to solve the k-Min-Sum-Radii problem.',
    'long_description': '[![Build Status](https://github.com/algo-hhu/k-min-sum-radii/actions/workflows/mypy-flake-test.yml/badge.svg)](https://github.com/algo-hhu/k-min-sum-radii/actions)\n[![License: GPL v3](https://img.shields.io/badge/License-GPLv3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0)\n[![Supported Python version](https://img.shields.io/badge/python-3.10+-blue.svg)](https://www.python.org/downloads/release/python-3100/)\n[![Stable Version](https://img.shields.io/pypi/v/kmsr?label=stable)](https://pypi.org/project/kmsr/)\n\n# K-Min-Sum-Radii\n\nkMSR provides various implementations to solve the k-Min-Sum-Radii problem. The k-Min-Sum-Radii problem is a clustering problem that aims to minimize the sum of the radii of the clusters. Given a set of points, the aim is to find $k$ balls such that the the sum of the radii of the balls is minimized. The package provides the following algorithms:\n\n- `fpt-heuristic`: The algorithm is an heuristic version of what is described in [this paper](https://doi.org/10.1007/978-3-031-49815-2_9). In practice, this works well for clusters that are not too separated. The algorithm uses the parameters `epsilon`, `n_u`, and `num_test_radii` to control the trade-off between the quality of the solution and the runtime. Increase `n_u` for a more accurate solution, which will however require more computation time.\n- `heuristic`: The algorithm is a simple heuristic that explores all possible combinations for the first cluster, and then selects the remaining centers as the points farthest from the radius of the first cluster. This algorithm works well in practice but it is not practical for large datasets.\n- `gonzalez`: This is the standard [Gonzalez algorithm](https://doi.org/10.1016/0304-3975(85)90224-5) for $k$-center.\n- `kmeans`: This is the [k-means++ algorithm](https://doi.org/10.5555/1283383.1283494) for $k$-means.\n\nAlthough the last two algorithms are not specifically designed for the k-Min-Sum-Radii problem, they serve as useful baselines for comparing the performance of other algorithms. Additionally, an optimization has been integrated into all algorithms: intersecting balls are merged at the end, resulting in a more cost-effective solution.\n\nFor computing the minimum enclosing circle of a set of points we use the [miniball](https://people.inf.ethz.ch/gaertner/subdir/software/miniball.html) [header](https://github.com/algo-hhu/k-min-sum-radii/blob/main/kmsr/header/miniball.h) for C++. Their code has not been changed.\n\n**You can try kMSR out on our [Clustering Toolkit](https://clustering-toolkit.algo.hhu.de/K-Min_Sum_Radii_Clustering)!**\n\n\n## Installation\n\nWe **highly recommend** to install OpenMP. Parts of the code are parallelized and will be much faster. However, on Windows and MacOS the code also works without OpenMP. Nonetheless, the code was written for Linux and will achieve the best performance there.\n\nOn Linux, you can use the following command:\n```bash\n# Alpine\nsudo apk add openmp-dev libgomp\n# Ubuntu\nsudo apt-get install libomp-dev libgomp1\n# Debian\nsudo apt-get install gcc libomp-dev libomp5 libgomp1\n# ArchLinux\nsudo pacman -S openmp\nln -s libomp.so libomp.so.5\n```\nWe have tested this on docker, so it might be a little different on your system.\n\nOn MacOS, you can use the following command:\n```bash\n brew install llvm libomp\n```\nHowever, it might be that MacOS does not find the installed library. In [`build_extension.py`](./build_extension.py), the paths are set manually. If it does not work for you, please clone the repository and run `poetry build`. You will see a message in red if your OpenMP is not found.\n\nThen, you can install the package via pip:\n```bash\npip install kmsr\n```\n\n## Example\n\n```python\nfrom kmsr import KMSR\nfrom kmsr.plot import plot_multiple_results\nfrom sklearn.datasets import make_blobs\nfrom time import time\n\npoints, ground_truth = make_blobs(\n    n_samples=100,\n    n_features=2,\n    centers=2,\n    cluster_std=0.6,\n    shuffle=True,\n    random_state=42,\n)\n\nlabels = []\ncenters = []\nradii = []\ntitles = []\nfor algo in ["FPT-heuristic", "Heuristic", "Gonzalez", "KMeans"]:\n    kmsr = KMSR(\n        n_clusters=5,\n        algorithm=algo,\n        epsilon=0.5,\n        n_u=10000,\n        n_test_radii=10,\n        random_state=42,\n    )\n    start = time()\n    kmsr.fit(points)\n    end = time() - start\n    labels.append(kmsr.labels_)\n    centers.append(kmsr.cluster_centers_)\n    radii.append(kmsr.cluster_radii_)\n    titles.append(f"{algo}: {sum(kmsr.cluster_radii_):.3f}, Time: {end:.3f}s")\n\nplot_multiple_results(\n    points,\n    clusters=labels,\n    centers=centers,\n    radii=radii,\n    title=titles,\n)\n```\n\n<p align="center">\n  <img src="https://raw.githubusercontent.com/algo-hhu/k-min-sum-radii/main/images/example.png" alt="Comparison of the Different Methods for kMSR"/>\n</p>\n\n## Development\n\nInstall [poetry](https://python-poetry.org/docs/#installation)\n```bash\ncurl -sSL https://install.python-poetry.org | python3 -\n```\n\nInstall clang\n```bash\nsudo apt-get install clang\n```\n\nSet clang variables\n```bash\nexport CXX=/usr/bin/clang++\nexport CC=/usr/bin/clang\n```\n\nInstall the package\n```bash\npoetry install\n```\n\nIf the installation does not work and you do not see the C++ output, you can build the package to see the stack trace\n```bash\npoetry build\n```\n\nRun the tests\n```bash\npoetry run python -m unittest discover tests -v\n```\n\n## Citation\n\nIf you use this code, please cite the following bachelor thesis:\n\n```\nN. Lenßen, "Experimentelle Analyse von Min-Sum-Radii Approximationsalgorithmen". Bachelorarbeit, Heinrich-Heine-Universität Düsseldorf, 2024.\n```\n\nMoreover, depending on the selection of the `algorithm` parameter, you should also cite the [following paper](https://doi.org/10.1007/978-3-031-49815-2_9) for `algorithm=\'fpt-heuristic\'`:\n\n```\nL. Drexler, A. Hennes, A. Lahiri, M. Schmidt, and J. Wargalla, "Approximating Fair K-Min-Sum-Radii in Euclidean Space," in Lecture notes in computer science, 2023, pp. 119–133. doi: 10.1007/978-3-031-49815-2_9.\n```\n\nthe [following paper](https://doi.org/10.1016/0304-3975(85)90224-5) for `algorithm=\'gonzalez\'`:\n\n```\nT. F. Gonzalez, "Clustering to minimize the maximum intercluster distance," Theoretical Computer Science, vol. 38, pp. 293–306, Jan. 1985, doi: 10.1016/0304-3975(85)90224-5.\n```\n\nand the [following paper](https://doi.org/10.5555/1283383.1283494) for `algorithm=\'kmeans\'`:\n\n```\nD. Arthur and S. Vassilvitskii, "k-means++: the advantages of careful seeding," Symposium on Discrete Algorithms, pp. 1027–1035, Jan. 2007, doi: 10.5555/1283383.1283494.\n```\n',
    'author': 'Niklas Lenssen',
    'author_email': 'niklas.lenssen@hhu.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}
from build_extension import *
build(setup_kwargs)

setup(**setup_kwargs)
